#include <iostream>

#include "texas.h"
#include <vector>
#include <cassert>
#include <cstring>
#include <algorithm>
#include <random>
#include <set>

namespace {
    enum class HandType {
        StraightFlush = 8,
        Four = 7,
        FullHouse = 6,
        Flush = 5,
        Straight = 4,
        Three = 3,
        TwoPair = 2,
        Pair = 1,
        HighCard = 0
    };

    struct Score{
        HandType w;
        int hash_value;
        Score(HandType _w, std::vector<int>& A) {
            w = _w; hash_value = 0;
            for (int i = 0; i < A.size(); ++i) {
                hash_value = hash_value * 20 + A[i];
            }
        }
        Score(){}
    };
    int operator < (const Score& k1, const Score& k2){
        return k1.w < k2.w || (k1.w == k2.w && k1.hash_value < k2.hash_value);
    }

    struct CardCmp {
        bool operator () (const Card& x, const Card& y) const {
            return x.value < y.value || (x.value == y.value && x.color < y.color);
        }
    };

    bool is_card_used[14][5];
    int used_num[14];
    std::mt19937 rd(0);

    bool checkStraightFlush(Score& res) {
        for (int c = 0; c < 4; ++c) {
            for (int i = 9; i >= 0; --i) {
                int flag = 0;
                for (int j = 0; j <= 4; ++j)
                    if (!is_card_used[i + j][c]) {
                        flag = 1;
                        break;
                    }
                if (flag) continue;
                std::vector<int> A;
                for (int j = 4; j >= 0; j--) A.push_back(i + j);
                res = Score(HandType::StraightFlush, A);
                return true;
            }
        }
        return false;
    }

    bool checkFour(Score& res) {
        for (int i = 13; i; i--)
            if (used_num[i] == 4) {
                std::vector<int> A;
                for (int k = 0; k < 4; k++) A.push_back(i);
                for (int j = 13; j; j--)
                    if (i != j && used_num[j]) {
                        A.push_back(j);
                        res = Score(HandType::Four, A);
                        return true;
                    }
            }
        return false;
    }

    bool checkFullHouse(Score &res) {
        for (int i = 13; i; i--)
            if (used_num[i] == 3) {
                std::vector<int> A;
                for (int k = 0; k < 3; k++) A.push_back(i);
                for (int j = 13; j; j--)
                    if (i != j && used_num[j] >= 2) {
                        for (int k = 0; k < 2; k++) A.push_back(j);
                        res = Score(HandType::FullHouse, A);
                        return true;
                    }
            }
        return false;
    }

    bool checkFlush(Score &res) {
        for (int c = 0; c < 4; c++) {
            std::vector<int> A;
            for (int i = 13; i; i--)
                if (is_card_used[i][c]) {
                    A.push_back(i);
                    if (A.size() == 5) {
                        res = Score(HandType::Flush, A);
                        return true;
                    }
                }
        }
        return false;
    }

    bool checkStraight(Score &res) {
        for (int i = 9; i >= 0; i--) {
            int flag = 0;
            for (int j = 4; j >= 0; j--)
                if (!used_num[i + j]) {
                    flag = 1;
                    break;
                }
            if (flag) continue;
            std::vector<int> A;
            for (int j = 4; j >= 0; j--)
                A.push_back(i + j);
            res = Score(HandType::Straight, A);
            return true;
        }
        return false;
    }

    bool checkThree(Score &res) {
        for (int i = 13; i; i--)
            if (used_num[i] >= 3) {
                std::vector<int> A;
                for (int k = 0; k < 3; k++) A.push_back(i);
                for (int j = 13; j; j--)
                    if (i != j && used_num[j]) {
                        A.push_back(j);
                        if (A.size() == 5) {
                            res = Score(HandType::Three, A);
                            return true;
                        }
                    }
            }
        return false;
    }

    bool checkTwoPair(Score &res) {
        for (int i = 13; i; i--)
            if (used_num[i] >= 2) {
                std::vector<int> A;
                for (int k = 0; k < 2; k++) A.push_back(i);
                for (int j = i - 1; j; j--)
                    if (used_num[j] >= 2) {
                        for (int k = 0; k < 2; k++) A.push_back(j);
                        for (int x = 13; x; x--)
                            if (used_num[x] && x != i && x != j) {
                                A.push_back(x);
                                res = Score(HandType::TwoPair, A);
                                return true;
                            }
                    }
            }
        return false;
    }

    bool checkPair(Score &res) {
        for (int i = 13; i; i--)
            if (used_num[i] >= 2) {
                std::vector<int> A;
                for (int k = 0; k < 2; k++) A.push_back(i);
                for (int j = 13; j; j--)
                    if (used_num[j] && j != i) {
                        A.push_back(j);
                        if (A.size() == 5) {
                            res = Score(HandType::Pair, A);
                            return true;
                        }
                    }
            }
        return false;
    }

    void getHighCard(Score &res) {
        std::vector<int> A;
        for (int i = 13; i; i--)
            if (used_num[i]) {
                A.push_back(i);
                if (A.size() == 5) {
                    res = Score(HandType::HighCard, A);
                    return;
                }
            }
    }

    Score getScoreForHand(std::vector<Card>& A) {
        memset(is_card_used, 0x00, sizeof is_card_used);
        memset(used_num, 0x00, sizeof used_num);
        for (int i = 0; i < A.size(); ++i) {
            is_card_used[A[i].value][A[i].color] = true;
            ++used_num[A[i].value];
        }
        used_num[0] = used_num[13];
        for (int i = 0; i <= 4; i++) {
            is_card_used[0][i] = is_card_used[13][i];
        }
        Score res;
        if (checkStraightFlush(res)) return res;
        if (checkFour(res)) return res;
        if (checkFullHouse(res)) return res;
        if (checkFlush(res)) return res;
        if (checkStraight(res)) return res;
        if (checkThree(res)) return res;
        if (checkTwoPair(res)) return res;
        if (checkPair(res)) return res;
        getHighCard(res);
        return res;
    }

    bool isValidCard(const Card& card) {
        return 1 <= card.value && card.value <= 13 && 0 <= card.color && card.color <= 3;
    }

    std::string myToString(int w) {
    	static char ch[100];
    	sprintf(ch, "%d", w);
    	return ch;
    }

    std::string cardToString(const Card& c) {
        return myToString(c.color) + "@" + myToString(c.value);
    }

    void assertCompleteState(const State& s) {
        static std::set<Card, CardCmp> S;
        S.clear();
        for (int i = 0; i < 2; ++i) {
            assert(isValidCard(s.alice[i]) && S.find(s.alice[i]) == S.end());
            S.insert(s.alice[i]);
            assert(isValidCard(s.bob[i]) && S.find(s.bob[i]) == S.end());
            S.insert(s.bob[i]);
        }
        for (int i = 0; i < 5; ++i) {
            assert(isValidCard(s.community[i]) && S.find(s.community[i]) == S.end());
            S.insert(s.community[i]);
        }
    }
}

int getResult(const State& s) {
    std::vector<Card> alice, bob;
    assertCompleteState(s);
    for (int i = 0; i < 2; ++i) {
        alice.push_back(s.alice[i]);
        bob.push_back(s.bob[i]);
    }
    for (int i = 0; i < 5; ++i) {
        alice.push_back(s.community[i]);
        bob.push_back(s.community[i]);
    }
    Score alice_score = getScoreForHand(alice);
    Score bob_score = getScoreForHand(bob);
    if (alice_score < bob_score) return -1;
    if (bob_score < alice_score) return 1;
    return 0;
}

std::pair<double, double> getRatesBySampling(const State& s, int t) {
    std::memset(is_card_used, 0x00, sizeof is_card_used);
    auto deal_with_card = [&](const Card& c) {
        if (isValidCard(c)) {
            assert(!is_card_used[c.value][c.color]);
            is_card_used[c.value][c.color] = true;
        }
    };
    for (int i = 0; i < 2; ++i) {
        deal_with_card(s.alice[i]);
        deal_with_card(s.bob[i]);
    }
    for (int i = 0; i < 5; ++i) {
        deal_with_card(s.community[i]);
    }
    std::vector<Card> remaining_card;
    for (int i = 1; i <= 13; ++i) {
        for (int j = 0; j < 4; ++j) {
            if (!is_card_used[i][j]) {
                remaining_card.emplace_back(j, i);
            }
        }
    }
    int win_num = 0, draw_num = 0;
    for (int _ = 0; _ < t; ++_) {
        State complete_s = s;
        std::shuffle(remaining_card.begin(), remaining_card.end(), std::default_random_engine(rd()));
        int ind = 0;
        for (int i = 0; i < 2; ++i) {
            if (!isValidCard(s.alice[i])) {
                complete_s.alice[i] = remaining_card[ind++];
            }
            if (!isValidCard(s.bob[i])) {
                complete_s.bob[i] = remaining_card[ind++];
            }
        }
        for (int i = 0; i < 5; ++i) {
            if (!isValidCard(s.community[i])) {
                complete_s.community[i] = remaining_card[ind++];
            }
        }
        auto result = getResult(complete_s);
        if (result == 1) {
            ++win_num;
        } else if (result == 0) {
            ++draw_num;
        }
    }
    return std::make_pair(1.0 * win_num / t, 1.0 * draw_num / t);
}

namespace {

    std::pair<ActionType, int> BobAction(const State& s, int rise, int pool_value) {
        std::pair<double, double> rates = getRatesBySampling(s, 100);
        double expected_call = -rates.first * rise + rates.second * pool_value / 2.0 + (1 - rates.first - rates.second) * (pool_value + rise);
        if (expected_call > 0) return std::make_pair(ActionType::CALL, 0);
        return std::make_pair(ActionType::FOLD, 0);
    }

    State getPartialState(const State& complete_state, int round, int player) {
        int num = 0;
        State s = complete_state;
        if (round == 0) num = 0; else num = 2 + round;
        for (int i = num; i < 5; ++i) {
            s.community[i] = Card();
        }
        if (player == 0) {
            for (int i = 0; i < 2; ++i) {
                s.bob[i] = Card();
            }
        } else {
            for (int i = 0; i < 2; ++i) {
                s.alice[i] = Card();
            }
        }
        return s;
    }

    const int KHand = 100;

    int runGame() {
        int pool_value = 10, hand_value = KHand;
        std::vector<Card> full_card_pool;
        for (int i = 1; i <= 13; ++i) {
            for (int j = 0; j < 4; ++j) {
                full_card_pool.emplace_back(j, i);
            }
        }
        std::shuffle(full_card_pool.begin(), full_card_pool.end(), std::default_random_engine(rd()));
        State complete_state;
        for (int i = 0; i < 2; ++i) {
            complete_state.alice[i] = full_card_pool[i];
            complete_state.bob[i] = full_card_pool[i + 2];
        }
        for (int i = 0; i < 5; ++i) {
            complete_state.community[i] = full_card_pool[i + 4];
        }

        for (int i = 0; i <= 3; ++i) {
            auto alice_act = makeDecision(getPartialState(complete_state, i, 0));
            assert(alice_act.first != ActionType::CALL);
            if (alice_act.first == ActionType::FOLD) return hand_value;
            else if (alice_act.first == ActionType::RAISE) {
                assert(alice_act.second <= hand_value);
                auto bob_act = BobAction(getPartialState(complete_state, i, 1), alice_act.second, pool_value);
                assert(bob_act.first == ActionType::CALL || bob_act.first == ActionType::FOLD);
                if (bob_act.first == ActionType::CALL) {
                    pool_value += alice_act.second * 2;
                    hand_value -= alice_act.second;
                } else {
                    return hand_value + pool_value;
                }
            }
        }

        int result = getResult(complete_state);
        if (result == 1) return hand_value + pool_value;
        else if (result == 0) return hand_value + pool_value / 2;
        else return hand_value;
    }
    const int KRoundNum = 50000;
}

int main() {
    int tot = 0;
    for (int t = 1; t < KRoundNum; ++t) {
        tot += runGame() - KHand;
    }
    printf("%.6lf\n", 1.0 * tot / KRoundNum);
    return 0;
}
